"""Base classes and registry for notebook figure export.

Copyright 2008-2012 Research Foundation State University of New York 
This file is part of QUB Express.                                          

QUB Express is free software; you can redistribute it and/or modify          
it under the terms of the GNU General Public License as published by 
the Free Software Foundation, either version 3 of the License, or    
(at your option) any later version.                                  

QUB Express is distributed in the hope that it will be useful,               
but WITHOUT ANY WARRANTY; without even the implied warranty of       
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
GNU General Public License for more details.                         

You should have received a copy of the GNU General Public License,   
named LICENSE.txt, in the QUB Express program directory.  If not, see        
<http://www.gnu.org/licenses/>.                                      

"""

from cStringIO import StringIO
import os
import sys
import traceback
import qubx.accept
import qubx.pyenv
import qubx.settings
import qubx.tree
from qubx.util_types import Anon, WeakEvent, Reffer

class NbItem(Anon):
    """The simplest type of notebook item; this class has the minimal properties and methods to implement the notebook protocol:

    mnu_caption: text for the notebook menu, e.g. "Chart"
    global_name: How to find this item from the global namespace, e.g. QubX.AmpHist.plots[0].nbChart
                 (for script recording)
    to_txt: function() -> string, returns plain-text representation of this item

Subclasses may present additional information, e.g. pictures or data series, but should
also provide to_txt for compatibility with console output.
"""
    def __init__(self, mnu_caption, global_name, to_txt, *args, **kw):
        Anon.__init__(self, mnu_caption=mnu_caption, global_name=global_name, to_txt=to_txt, *args, **kw)
    def __str__(self):
        return self.to_txt()

class NbItems(NbItem):
    """A bundle of related notebook items; e.g. "All Charts".  Each item in items is an instance of NbItem or a subclass.

    >>> if isinstance(myNbItem, NbItems):
    ...     for item in myNbItem:
    ...         print item

"""
    def __init__(self, mnu_caption, global_name, *items):
        NbItem.__init__(self, mnu_caption, global_name, self.__to_txt, items=list(items))
    def iteritems(self):
        for item in self.items:
            if item:
                yield item
    def __iter__(self):
        return self.iteritems()
    def __to_txt(self):
        buf = StringIO()
        for item in self:
            buf.write(item.to_txt())
        return buf.getvalue()

class NbText(NbItem):
    """A simple notebook item with static text."""
    def __init__(self, msg):
        NbItem.__init__(self, "Text", "", lambda: msg)

class NbDynText(NbItem):
    """A simple notebook item with text that's recomputed by the get_msg function whenever it's needed."""
    def __init__(self, mnu_caption, global_name, get_msg):
        NbItem.__init__(self, mnu_caption, global_name, get_msg)

class NbTable(NbItem):
    """A notebook item with tabular data. Subclasses NbChart and NbTrace define series using this underlying table of data.

    get_caption: function() -> str, returns caption for the table
    get_shape: function() -> (rows, columns), returns dimensions
    get_headers: function() -> list of str, returns column headers
    get_row: function(r) -> list, returns a row of data
    get_col: function(c) -> list, returns a column of data
             (When creating an NbTable, provide either get_row or get_col and the other will be auto-generated;
              when using an NbTable, both get_row and get_col are available.)
    get_col_format: function(c) -> (function(val) -> str), returns a function which formats values from column c as strings (by default, the str function)
    get_type: function() -> type; returns e.g. float or str
"""
    def __init__(self, mnu_caption, global_name, get_caption, get_shape, get_headers, get_row=None, get_col=None, get_col_format=None, get_type=None,
                 **kw):
        the_get_row = get_row or self.__get_row
        the_get_col = get_col or self.__get_col
        the_get_col_format = get_col_format or self.__get_col_format
        the_get_type = get_type or self.__get_type
        NbItem.__init__(self, mnu_caption=mnu_caption, global_name=global_name,
                        get_caption=get_caption, to_txt=self.__to_txt, get_shape=get_shape, get_headers=get_headers,
                      get_row=the_get_row, get_col=the_get_col, get_col_format=the_get_col_format,
                      get_type=the_get_type, **kw)
    def __get_type(self):
        return str
    def __get_col_format(self, c):
        return str
    def __get_col(self, c):
        Nr, Nc = self.get_shape()
        return [self.get_row(r)[c] for r in xrange(Nr)]
    def __get_row(self, r):
        Nr, Nc = self.get_shape()
        return [(len(col) > r) and col[r] or 0.0 for col in [self.get_col(c) for c in xrange(Nc)]]
    def __to_txt(self):
        buf = StringIO()
        buf.write('\t'.join(self.get_headers()))
        buf.write('\n')
        Nr, Nc = self.get_shape()
        col_format = [self.get_col_format(c) for c in xrange(Nc)]
        if self.__get_row != self.get_row: # rows provided directly:
            for r in xrange(Nr):
                buf.write('\t'.join(col_format[c](x) for c,x in enumerate(self.get_row(r))))
                buf.write('\n')
        else:
            cols = [self.get_col(c) for c in xrange(Nc)]
            for r in xrange(Nr):
                buf.write('\t'.join((len(cols[c]) > r) and col_format[c](cols[c][r]) or "0.0" for c in xrange(Nc)))
                buf.write('\n')
        return buf.getvalue()

def NbSubTable(name='SubTable', fields=[]):
    "Returns a function(qubx.table.Table), which returns an NbTable with only the specified columns."
    def make_nb(table):
        return qubx.notebook.NbTable(name, table.global_name and ('%s.notebook[%s]'%(table.global_name, repr(name))) or '',
                                     get_caption=table.nb_get_caption,
                                     get_shape=lambda: (table.nb_get_shape()[0], len([f for f in fields if f in table.fields])),
                                     get_headers=lambda: [f for f in fields if f in table.fields],
                                     get_row=lambda r: [table[r,f] for f in fields if f in table.fields],
                                     get_col_format=lambda c: table.nb_get_col_format(table.fields.index([f for f in fields if f in table.fields][c])),
                                     get_type=lambda: table.nb_get_type([f for f in fields if f in table.fields]))
    make_nb.name = name
    return make_nb

class NbStaticTable(NbTable):
    """A notebook item representing a specific array."""
    def __init__(self, arr, caption="", header="", format=str, as_column=True):
        kw = {}
        if as_column:
            rows, cols = len(arr), 1
            kw["get_col"] = lambda i: arr
        else:
            rows, cols = 1, len(arr)
            kw["get_row"] = lambda i: arr
        fmt = qubx.accept.acceptFormat(format)
        kw["get_col_format"] = lambda i: fmt
        if len(arr) and hasattr(arr[0], "replace"):
            kw["get_type"] = lambda: str
        else:
            kw["get_type"] = lambda: float
        NbTable.__init__(self, "Table", "", lambda: caption, lambda: (rows, cols), lambda: [header]*cols, **kw)

class NbStaticArray(NbTable):
    """A notebook item representing a specific 1d or 2d numpy.ndarray."""
    def __init__(self, arr, caption="", headers=[], format=str):
        kw = {}
        if len(arr.shape) == 2:
            rows, cols = arr.shape
            kw["get_col"] = lambda i: arr[:,i].flatten()
            kw["get_row"] = lambda i: arr[i,:].flatten()
        else:
            rows, cols = 1, len(arr)
            kw["get_col"] = lambda i: arr
        fmt = qubx.accept.acceptFormat(format)
        kw["get_col_format"] = lambda i: fmt
        kw["get_type"] = lambda: float
        hdrs = headers + (cols - len(headers)) * [""]
        NbTable.__init__(self, "Table", "", lambda: caption, lambda: (rows, cols), lambda: hdrs, **kw)



DOTS = 0
LINES = 1
HISTOGRAM = 2

def c2h(c):
    h = '%x'%c
    if len(h) == 1:
        return '0%s'%h
    return h
def rgb_to_hex(rgb):
    return '#%s' % ''.join(c2h(255*x) for x in rgb[:3])

class NbChartSeries(Anon):
    """Properties of one series in a chart item:

    xcol: index of the X column
    ycol: index of the Y column
    first_row: this series may start on a row other than 0
    nrow: and consist of fewer rows than the entire table
    ser_type: one of DOTS, LINES, or HISTOGRAM = (0, 1, 2)
    color_rgb: (r, g, b) with each coordinate between 0.0 and 1.0
    weight: thickness of line, between 0.0 and 1.0
"""
    def __init__(self, xcol, ycol, first_row=0, nrow=-1, ser_type=DOTS, color_rgb=(0,0,0), weight=1.0):
        Anon.__init__(self, xcol=xcol, ycol=ycol, first_row=first_row, nrow=nrow,
                      ser_type=ser_type, color_rgb=color_rgb, color_hex=rgb_to_hex(color_rgb), weight=weight)

class NbChart(NbTable):
    """Notebook item representing a 2D plot.  In addition to the data defined by its L{NbTable} properties, it defines
L{NbChartSeries} to lay out dots, lines and histograms using the table data.

    get_xlabel: function() -> string, caption for the x axis
    get_ylabel: function() -> string, caption for the y axis
    get_series: function() -> list of L{NbChartSeries}
    get_colors: function() -> list of (r,g,b) (or None), optionally defining a palette
    get_color_indices: function -> list of int (or None), optional colors for the first series
"""
    def __init__(self, mnu_caption, global_name, get_caption, get_shape, get_headers, get_xlabel=lambda: "", get_ylabel=lambda: "",
                 get_series=lambda: [], get_colors=lambda:None, get_color_indices=lambda:None, *args, **kw):
        kw['get_xlabel'] = get_xlabel
        kw['get_ylabel'] = get_ylabel
        kw['get_series'] = get_series
        kw['get_colors'] = get_colors
        kw['get_color_indices'] = get_color_indices
        NbTable.__init__(self, mnu_caption, global_name, get_caption, get_shape, get_headers, *args, **kw)

class NbTrace(NbChart):
    """Notebook item representing sampled data from one or more signals.

    get_trace_series: function() -> list of L{NbChartSeries}
"""
    
    def __init__(self, mnu_caption, global_name, get_caption, get_shape, get_headers, get_trace_series=lambda: [], *args, **kw):
        kw['get_trace_series'] = get_trace_series
        NbChart.__init__(self, mnu_caption, global_name, get_caption, get_shape, get_headers, *args, **kw)


### NbPicture in qubx.notebookGTK


class NbTarget(object): # inheritance optional; just implement the methods
    """Base class for objects which receive notebook items. Multiple NbTargets can be registered by name with the NbController qubx.notebook.Notebook.
Targets need not inherit from this class, but must implement these properties and methods:

    supports: list of subclasses of NbItem which this target can handle, e.g. [NbText, NbDynText, NbTable]
    nb_init(): called when the target is (re-) activated
    nb_send(item): acts on the L{NbItem}
"""
    def __init__(self):
        self.supports = [] # e.g. [NbText, NbDynText, NbTable, NbChart, NbTrace, NbPicture]
    def nb_init(self):
        pass
    def nb_send(self, item):
        pass

class NbTargetRec(Anon):
    """Info about one registered notebook target:

    target_id: short string to identify target in code
    caption: user-visible name of target
    menu_caption: shortened user-visible name, for menus
    controls: gtk.Widget or None; additional gui elements
    target: L{NbTarget}
"""
    pass # target_id, caption, menu_caption, controls, target

class NbAutoRec(Anon):
    """Info about one automatic notebook item:

    auto_id: string to programmatically identify this item and the occasion for automatically generating it
    mnu_caption: user-visible description
    active: False if this item is disabled (un-checked by user)
"""
    pass # auto_id, mnu_caption, active

class NbController(object):
    """The global instance of this type, qubx.notebook.Notebook, is created by qubx.notebook.Init(),
and keeps registries of the various output methods (L{NbTarget}s) and automatically-triggered L{NbItem}s.

    @ivar auto: dict(auto_id -> L{NbAutoRec})
    @ivar targets: list of L{NbTargetRec}
    @ivar target_ix: index of user's choice in targets
    @ivar target_id: chosen target's target_id
    @ivar target: chosen L{NbTarget}
    @ivar OnAddAuto: L{WeakEvent}(index, L{NbAutoRec})
    @ivar OnChangeAuto: L{WeakEvent}(index, L{NbAutoRec}) e.g. the user has toggled "active"
    @ivar OnAddTarget: L{WeakEvent}(index)
    @ivar OnRemovingTarget: L{WeakEvent}(index)
    @ivar OnChangeTargetIx: L{WeakEvent}(index)
"""
    def __init__(self):
        self.__ref = Reffer()
        self.auto = {} # auto_id : NbAutoRec
        self.targets = [] # NbTargetRec
        self.__target_ix = -1
        self.OnAddAuto = WeakEvent()
        self.OnChangeAuto = WeakEvent()
        self.OnAddTarget = WeakEvent()
        self.OnRemovingTarget = WeakEvent()
        self.OnChangeTargetIx = WeakEvent()
        self.props = qubx.tree.Node() # defer until targets are registered
        self.settings_auto = qubx.settings.SettingsMgr['qubx.notebook.NbController.auto']
        self.settings_auto.OnSet = self.__ref(self.__onSetPropsAuto)
        self.props_auto = self.settings_auto.active
    def init_settings(self):
        self.settings = qubx.settings.SettingsMgr['qubx.notebook.NbController']
        have_target = not self.settings.active.find('target_id').isNull
        self.settings.OnSet = self.__ref(self.__onSetProps)
        self.__onSetProps(self.settings, self.settings.active)
        self.settings.active = self.props
        return have_target
    def set_target_ix(self, x):
        if x == self.__target_ix:
            return
        self.__target_ix = x
        self.props['target_id'].data = self.targets[x].target_id
        self.__finish_set_target_ix()
    def __finish_set_target_ix(self):
        if not qubx.pyenv.env.globals['QUBX_READY']:
            qubx.pyenv.env.call_later(self.__finish_set_target_ix)
            return
        self.targets[self.__target_ix].target.nb_init()
        self.OnChangeTargetIx(self.__target_ix)
    target_ix = property(lambda self: self.__target_ix, lambda self, x: self.set_target_ix(x))
    def get_target_id(self):
        if self.__target_ix < 0:
            return ""
        return self.targets[self.__target_ix].target_id
    def set_target_id(self, x):
        if (self.__target_ix >= 0) and (x == self.targets[self.__target_ix].target_id):
            return
        for i, t in enumerate(self.targets):
            if t.target_id == x:
                self.set_target_ix(i)
                break
    target_id = property(lambda self: self.get_target_id(), lambda self, x: self.set_target_id(x))
    target = property(lambda self: self.targets[self.__target_ix])
    def list_target_ids(self):
        """Returns all registered target_id values."""
        return [target.target_id for target in self.targets]
    def __onSetProps(self, settings, updates):
        if updates.find('target_id').data:
            self.target_id = str(updates['target_id'].data)
            self.props['target_id'].data = self.target_id
    def __onSetPropsAuto(self, settings, updates):
        for auto in qubx.tree.children(updates):
            nm = auto.name
            if nm and (nm in self.auto) and auto.data:
                active = bool(auto.data[0])
                if active != self.auto[nm].active:
                    self.auto[nm].active = active
                    self.props_auto[nm].data = active
                    self.OnChangeAuto(sorted(self.auto.keys()).index(nm), self.auto[nm])
    def register_target(self, target, target_id, caption, menu_caption=None, controls=None):
        the_menu_caption = menu_caption
        if the_menu_caption is None:
            the_menu_caption = caption
        ix = -1
        for i,t in enumerate(self.targets):
            if t.target_id == target_id:
                ix = i
                break
        if ix >= 0:
            self.OnRemovingTarget(ix)
            del self.targets[ix]
        ix = len(self.targets)
        self.targets.append(NbTargetRec(target=target, target_id=target_id, caption=caption,
                                        menu_caption=the_menu_caption, controls=controls))
        self.OnAddTarget(ix)
        if ix == 0:
            self.target_ix = ix
    def register_auto(self, auto_id, mnu_caption, default=False):
        """Declares that automatic items will be generated with this auto_id and caption.  With default=False, the user will have to enable it in the Notebook tab."""
        preexist = auto_id in self.auto
        prop = self.props_auto[auto_id]
        if prop.data:
            active = bool(prop.data[0])
        else:
            active = default
            prop.data = active
        self.auto[auto_id] = NbAutoRec(auto_id=auto_id, mnu_caption=mnu_caption, active=active)
        ix = sorted(self.auto.keys()).index(auto_id)
        if preexist:
            self.OnChangeAuto(ix, self.auto[auto_id])
        else:
            self.OnAddAuto(ix, self.auto[auto_id])
    def set_auto_active(self, auto_id, active):
        self.auto[auto_id].active = active
        self.props_auto[auto_id].data = int(active)
        self.OnChangeAuto(sorted(self.auto.keys()).index(auto_id), self.auto[auto_id])
    def send(self, item, target_ix=None, auto_id=None, target_id=None):
        """Writes an L{NbItem} to the chosen L{NbTarget}.

    @param target_ix: index of target in targets, or None to use the user's choice
    @param target_id: if (target_ix is None), picks a target by its target_id
    @param auto_id: for items generated automatically rather than by explicit user or script request; will check if auto item is active before emitting
"""
        if auto_id and (auto_id in self.auto) and not self.auto[auto_id].active:
            return
        if not qubx.pyenv.env.globals['QUBX_READY']:
            if qubx.pyenv.env.globals['DEBUG']:
                print 'dropped early notebook item:'
                print item.to_txt()
            return
        ix = None
        if not (target_ix is None):
            ix = target_ix
        elif not (target_id is None):
            for i, t in enumerate(self.targets):
                if t.target_id == target_id:
                    ix = i
                    break
        if ix is None:
            ix = self.__target_ix

        try:
            if auto_id is None:
                targeted = not ((target_ix is None) and (target_id is None))
                if isinstance(item, NbText):
                    qubx.pyenv.env.OnScriptable('Notebook.send(qubx.notebook.NbText("""%s"""%s))' %
                                                (item.to_txt(),
                                                 targeted and (", target_id='%s'" % self.targets[ix].target_id) or ""))
                elif item.global_name:
                    qubx.pyenv.env.scriptable_if_matching("Notebook.send(%s%s)" %
                                                          (item.global_name,
                                                           targeted and (", target_id='%s'" % self.targets[ix].target_id) or ""),
                                                          [(item.global_name, item)])
                
            self.targets[ix].target.nb_send(item)
        except:
            traceback.print_exc()

class NbTarget_Stdout(object):
    """Basic type of notebook output which prints everything to the stdout console."""
    def __init__(self):
        self.supports = [NbItem] # [NbText, NbDynText, NbTable, NbChart, NbTrace]
    def nb_init(self):
        pass
    def nb_send(self, item):
        print item.to_txt()


def Init():
    """Sets up qubx.notebook.Notebook as the global L{NbController} object."""
    global Notebook
    Notebook = NbController()

def RegisterTargets():
    """Registers the target types defined in this module."""
    Notebook.register_target(NbTarget_Stdout(), 'stdout', 'Console', 'Write to console:', None)
    Notebook.register_target(NbTarget(), 'none', 'Discard', '', None)
